"""
Update time-step, depending on the flow field.
See :ref:`adaptive time_step` for details.
"""

from abc import ABCMeta, abstractmethod
from hysop.constants import HYSOP_REAL, StretchingCriteria, AdvectionCriteria
from hysop.tools.types import check_instance, first_not_None
from hysop.tools.decorators import debug
from hysop.tools.numpywrappers import npw
from hysop.core.graph.node_generator import ComputationalGraphNodeGenerator
from hysop.core.graph.computational_operator import ComputationalGraphOperator
from hysop.core.graph.graph import op_apply
from hysop.fields.continuous_field import Field
from hysop.parameters import ScalarParameter, TensorParameter

class TimestepCriteria(ComputationalGraphOperator):
    __metaclass__ = ABCMeta
    
    @debug
    def __init__(self, parameter, input_params, output_params, 
            dt_coeff=None, min_dt=None, max_dt=None, **kwds):
        """
        Initialize an AdaptiveTimeStep operator.

        Parameters
        ----------
        parameter: ScalarParameter
            Timestep parameter that will be updated.
        input_params: dict
            Input parameters used to compute criteria.
        output_params: dict
            Input parameters used to compute criteria.
        min_dt : float, optional
            Minimum value allowed for time step, defaults to 0.
        max_dt : float, optional
            Maximum value allowed for time step, defaults to 1e8.
        dt_coeff: float, optional
            Constant coefficient applied to resulting dt
            before resolving min and max values.
            Defaults to 1.
        kwds: dict
            Base class arguments.
        """
        check_instance(parameter, ScalarParameter)
        check_instance(input_params,  dict, keys=str)
        check_instance(output_params, dict, keys=str)
        assert (min_dt is None) or (min_dt > 0.0)
        assert (max_dt is None) or (max_dt > 0.0)
        assert (min_dt is None) or (max_dt is None) or (max_dt >= min_dt)
        assert (dt_coeff is None) or (dt_coeff > 0.0)
        assert parameter.name in output_params
        
        super(TimestepCriteria,self).__init__(input_params=input_params, 
                output_params=output_params, **kwds)

        self.min_dt   = 0.0 if (min_dt   is None) else min_dt
        self.max_dt   = 1e8 if (max_dt   is None) else max_dt
        self.dt_coeff = 1.0 if (dt_coeff is None) else dt_coeff 
        self.dt       = parameter

    @op_apply
    def apply(self, **kwds):
        dt = self.compute_criteria(**kwds)
        dt *= self.dt_coeff
        dt = npw.maximum(dt, self.min_dt)
        dt = npw.minimum(dt, self.max_dt)
        assert (dt > 0.0), 'negative or zero timestep encountered.'
        self.dt.set_value(dt)

    @abstractmethod
    def compute_criteria(**kwds):
        pass



class CflTimestepCriteria(TimestepCriteria):
    
    @debug
    def __init__(self, cfl, parameter,
            Finf=None, Fmin=None, Fmax=None, 
            dx=None, 
            name=None, pretty_name=None, 
            relative_velocities=None, **kwds):
        """
        Initialize a CflTimestepCriteria.

        Compute dt = cfl * Min_i(dx[i] / F[i].inf)

        cfl = (Vinf/dx)*dt

        Parameters
        ----------
        cfl: float
            CFL value used to compute timestep.
        Finf: TensorParameter
            A tensor parameter that contains Finf for every components.
            Can be specified instead of Fmin and Fmax:
                *Fmin will be set to -Finf
                *Fmax will be set to +Finf
        Fmin: TensorParameter
            A tensor parameter that contains Fmin for every components.
        Fmax: TensorParameter
            A tensor parameter that contains Fmax for every components.
        parameter: ScalarParameter
            The output parameter that will store the computed timestep.
        dx: tuple of float
            Space discretization, should be of size F.nb_components.
            If not given, will be extracted from Finf on setup.
        relative_velocities: array like of relative velocities, optional
            Specify relative velocities.
        kwds: dict
            Base class arguments.
        """
        assert (cfl > 0.0), 'negative cfl condition.'
        check_instance(cfl, float)
        if (Finf is None):
            check_instance(Fmin, TensorParameter)
            check_instance(Fmax, TensorParameter)
            assert Fmin.shape == Fmax.shape
            input_params={ Fmin.name: Fmin, Fmax.name: Fmax } 
            dtype = Fmin.dtype
            shape = Fmin.shape
            size  = Fmin.size
        else:
            check_instance(Finf, TensorParameter)
            msg='Cannot specify (Fmin,Fmax) and Finf at the same time.'
            assert (Fmin is None), msg
            assert (Fmax is None), msg
            input_params={ Finf.name: Finf } 
            dtype = Finf.dtype
            shape = Finf.shape
            size  = Finf.size

        check_instance(parameter, ScalarParameter)
        check_instance(dx, tuple, values=float, allow_none=True)
        check_instance(relative_velocities, (tuple, list), allow_none=True)

        name = first_not_None(name, 'CFL')
        pretty_name = first_not_None(pretty_name, name)
        super(CflTimestepCriteria,self).__init__(name=name, pretty_name=pretty_name,
                                                 input_params=input_params,
                                                 output_params={parameter.name: parameter},
                                                 parameter=parameter, **kwds)

        if (relative_velocities is None):
            relative_velocities = [(0,)*size]
        assert len(relative_velocities)>=1

        rv = ()
        for Vr in relative_velocities:
            Vr = npw.asarray(Vr, dtype=dtype)
            assert Vr.shape == shape
            rv += (Vr,)
        relative_velocities = rv

        self.cfl  = cfl
        self.Fmin = Fmin
        self.Fmax = Fmax
        self.Finf = Finf
        self.dx   = dx
        self.relative_velocities = relative_velocities

    def setup(self, **kwds):
        super(CflTimestepCriteria, self).setup(**kwds)
        if (self.dx is None):
            dx = None
            for attr in ('Finf', 'Fmin', 'Fmax'):
                attr = getattr(self, attr)
                if hasattr(attr, 'min_max_dfield'):
                    dx = attr.min_max_dfield.get_unique_attribute('mesh', '_mesh', '_space_step')[::-1]
            if (dx is None):
                msg='Could not extract dx from Fmin.'
                raise RuntimeError(msg)
            self.dx = dx

    def compute_criteria(self, **kwds):
        cfl = self.cfl
        dx = self.dx
        if (self.Finf is None):
            Fmin = self.Fmin()
            Fmax = self.Fmax()
        else:
            Finf = self.Finf()
            Fmin = -Finf
            Fmax = +Finf
        assert len(dx) == Fmin.size == Fmax.size
        assert len(self.relative_velocities)>=1

        dt = npw.inf
        for Vr in self.relative_velocities:
            Vmin = Fmin - Vr
            Vmax = Fmax - Vr
            Vinf = npw.maximum(npw.abs(Vmin), npw.abs(Vmax))
            if npw.all(npw.divide(Vinf, dx)==0):
                cdt = cfl*npw.inf
            else:
                cdt = cfl / npw.max(npw.divide(Vinf, dx))
            dt = min(dt, cdt)
        return dt
    
    def compute_cfl(self, dt):
        mdt = self.compute_criteria()
        return (dt / mdt) * self.cfl

class AdvectionTimestepCriteria(TimestepCriteria):

    @debug
    def __init__(self, lcfl, parameter, criteria,
                    Finf=None, gradFinf=None, 
                    name=None, pretty_name=None, **kwds):
        """
        Initialize a AdvectionTimestepCriteria.

        Compute a timestep criteria for advection:

        W_INF:       dt = lcfl / Max_i(|Wi|)
        GRAD_U:      dt = lcfl / Max_i(|dUi/dXi|)
        DEFORMATION: dt = lcfl / Max_i( Sum_j(|Sij|) )
                        where Sij is the deformation tensor
                        Sij = (dUi/dXj + dUj/dXi)/2

        where |dUi/dXj| = |gradF|_inf
              |Wi|      = |F|_inf

        Parameters
        ----------
        lcfl: float
            Lagrangian CFL constant.
        Finf: TensorParameter, optional
            A tensor parameter that contains |W|inf for every components.
            of the vorticity.
        gradFinf: TensorParameter
            A tensor parameter that contains |gradF|inf for every components 
            in every directions, ie. the inf. norm of the gradient of velocity.
        parameter: ScalarParameter
            The output parameter that will store the computed timestep.
        criteria: AdvectionCriteria
            The chosen criteria.
            Currently only W_INF, GRAD_U, and DEFORMATION is supported.
        kwds: dict
            Base class arguments.
        """
        assert (lcfl > 0.0), 'negative LCFL factor.'
        check_instance(lcfl, float)
        check_instance(Finf, TensorParameter, allow_none=True)
        check_instance(gradFinf, TensorParameter, allow_none=True)
        check_instance(parameter, ScalarParameter)
        check_instance(criteria, AdvectionCriteria)
        input_params = {}
        if (Finf is not None):
            assert Finf().ndim == 1, 'Finf should be a 1D tensor parameter.'
            input_params[Finf.name] = Finf
        if (gradFinf is not None):
            assert gradFinf().ndim == 2 , 'gradFinf should be a 2D tensor parameter.'
            input_params[gradFinf.name] = gradFinf

        name = first_not_None(name, 'LCFL')
        pretty_name = first_not_None(pretty_name, name)
        super(AdvectionTimestepCriteria, self).__init__(name=name, pretty_name=pretty_name,
                                                 input_params=input_params,
                                                 output_params={parameter.name: parameter},
                                                 parameter=parameter, **kwds)
        self.lcfl = lcfl
        self.Finf = Finf
        self.gradFinf = gradFinf
        self.criteria = criteria

    def compute_criteria(self, **kwds):
        criteria = self.criteria
        lcfl = self.lcfl
        if (criteria is AdvectionCriteria.W_INF):
            Finf = self.Finf()
            if npw.max(Finf)==0:
                return lcfl*npw.inf
            else:
                return lcfl / npw.max(Finf)
        elif (criteria is AdvectionCriteria.GRAD_U):
            gradFinf = self.gradFinf()
            if (gradFinf.ndim == 2):
                gradFinf = npw.diag(gradFinf) #extract diagonal
            return lcfl / npw.max(gradFinf)
        elif (criteria is AdvectionCriteria.DEFORMATION):
            gradFinf = self.gradFinf()
            gradFinf = (gradFinf + gradFinf.T)/2.0
            return lcfl / npw.max( gradFinf.sum(axis=0) )
        else:
            msg='Unsupported stretching criteria {}.'.format(criteria)
            raise RuntimeError(msg)

class StretchingTimestepCriteria(TimestepCriteria):
    
    @debug
    def __init__(self, gradFinf, parameter, 
                    cst=1.0, criteria=StretchingCriteria.GRAD_U,
                    name=None, pretty_name=None,
                    **kwds):
        """
        Initialize a StretchingTimestepCriteria.

        Compute dt = cst / Max_i( Sum_j( |dFi/dXj| ) )

        where |dFi/dXj| = |gradF|_inf_ij

        ie. dt = cst / npw.max( gradFinf.sum(axis=1) )

        Parameters
        ----------
        cst: float
            Factor used to compute timestep.
        gradFinf: TensorParameter
            A tensor parameter that contains |gradF|inf for every components in every directions.
        parameter: ScalarParameter
            The output parameter that will store the computed timestep.
        criteria: StretchingCriteria
            The chosen criteria.
            Currently only StretchingCriteria.GRAD_U is supported.
        kwds: dict
            Base class arguments.
        """
        assert (cst > 0.0), 'negative constant factor.'
        check_instance(cst, float)
        check_instance(gradFinf, TensorParameter)
        check_instance(parameter, ScalarParameter)
        check_instance(criteria, StretchingCriteria)
        assert gradFinf().ndim == 2, 'gradFinf should be a 2D tensor.'
        
        name = first_not_None(name, 'STRETCH')
        pretty_name = first_not_None(pretty_name, name)
        super(StretchingTimestepCriteria, self).__init__(name=name, pretty_name=pretty_name,
                                                 input_params={gradFinf.name: gradFinf}, 
                                                 output_params={parameter.name: parameter},
                                                 parameter=parameter, **kwds)
        self.cst  = cst
        self.gradFinf = gradFinf
        self.criteria = criteria

    def compute_criteria(self, **kwds):
        criteria = self.criteria
        if (criteria is StretchingCriteria.GRAD_U):
            gradFinf = self.gradFinf()
            return self.cst / npw.max( gradFinf.sum(axis=1) )
        else:
            msg='Unsupported stretching criteria {}.'.format(criteria)
            raise RuntimeError(msg)

class MergeTimeStepCriterias(TimestepCriteria):

    @debug
    def __init__(self, parameter, criterias, 
                    equivalent_CFL=None, cfl_criteria=None,
                    **kwds):
        check_instance(parameter, ScalarParameter)
        check_instance(criterias, dict, keys=str, values=TimestepCriteria)
        check_instance(equivalent_CFL, ScalarParameter, allow_none=True)
        output_params = { parameter.name: parameter }
        input_params = {}
        for criteria in criterias.values():
            input_params.update(criteria.output_params)
        
        assert not ((equivalent_CFL is not None) ^ (cfl_criteria is not None))
        self.equivalent_CFL = equivalent_CFL
        self.cfl_criteria = cfl_criteria
        
        super(MergeTimeStepCriterias, self).__init__(parameter=parameter,
                input_params=input_params,
                output_params=output_params,
                **kwds)

    def compute_criteria(self, **kwds):
        dt = min( p.value for p in self.input_params.values() )
        if (self.equivalent_CFL is not None):
            cfl =  self.cfl_criteria.compute_cfl(dt)
            self.equivalent_CFL.set_value(cfl)
        return dt
    
    @debug
    def apply(self, simulation, **kwds):
        assert simulation.dt is self.dt, 'Parameter mismatch between Simulation and AdaptiveTimeStep.'
        super(MergeTimeStepCriterias, self).apply(simulation=simulation, **kwds)

class AdaptiveTimeStep(ComputationalGraphNodeGenerator):
    """
    The adaptative Time Step is computed according
    to the following expression :

    dt = max(min(max_dt, dt_criterias), min_dt)
    """

    @debug
    def __init__(self, dt, min_dt=None, max_dt=None, dt_coeff=None, 
            equivalent_CFL=False, base_kwds=None, **kwds):
        """
        Initialize an AdaptiveTimeStep operator.

        Parameters
        ----------
        dt: hysop.parameter.scalar_parameter.ScalarParameter
            The parameter that will be updated.
        min_dt : double, optional
            Minimum value allowed for time step.
        max_dt : double, optional
            Maximum value allowed for time step.
        dt_coeff: double, optional
            Constant coefficient applied to resulting dt
            after resolving min and max values.
        param_name: str, optional
            Output dt parameter name (default is 'dt').
        base_kwds: dict
            Base kwds of this class.
        kwds : 
            Additional keywords arguments that will be passed 
            to MergeTimeStepCriterias.

        Notes
        -----
        * This operator has no effect on input variables.
        """
        base_kwds = first_not_None(base_kwds, {})
        super(AdaptiveTimeStep, self).__init__(
                candidate_input_tensors=(),
                candidate_output_tensors=(),
                **base_kwds)
        
        # tuple of criterias used to compute dt
        self.criterias = {}
        self.merge_kwds = { 'min_dt':     min_dt,
                            'max_dt':     max_dt,
                            'dt_coeff':   dt_coeff }
        self.merge_kwds.update(**kwds)
        self.equivalent_CFL = None
        self.cfl_criteria = None
        self.parameter = dt

    def push_cfl_criteria(self, cfl, Fmin=None, Fmax=None, dx=None,
            name=None, pretty_name=None, 
            param_name=None, param_pretty_name=None,
            parameter=None, quiet=False,
            relative_velocities=None,
            equivalent_CFL=None, **kwds):
        """
        See hysop.operator.adapt_timpestep.CflTimestepCriteria.
        """
        parameter = self._build_parameter(parameter=parameter, quiet=quiet,
                name=param_name, pretty_name=param_pretty_name, basename='cfl')
        criteria = CflTimestepCriteria(cfl=cfl, Fmin=Fmin, Fmax=Fmax, dx=dx, parameter=parameter, 
                                        name=name, pretty_name=pretty_name, 
                                        relative_velocities=relative_velocities, **kwds)
        self._push_criteria(parameter.name, criteria)
        
        if isinstance(equivalent_CFL, ScalarParameter):
            cfl_criteria = criteria
        elif (equivalent_CFL is True):
            equivalent_CFL = ScalarParameter(name='CFL*')
            cfl_criteria = criteria
        else:
            equivalent_CFL = None
            cfl_criteria = None
        self.equivalent_CFL = equivalent_CFL
        self.cfl_criteria = cfl_criteria
        return criteria.dt
    
    def push_advection_criteria(self, lcfl, criteria, Finf=None, gradFinf=None, 
            name=None, pretty_name=None, 
            param_name=None, param_pretty_name=None,
            parameter=None, quiet=False, **kwds):
        """
        See hysop.operator.adapt_timpestep.AdvectionTimestepCriteria.
        """
        parameter = self._build_parameter(parameter=parameter, quiet=quiet,
                name=param_name, pretty_name=param_pretty_name, basename='lcfl_{}'.format(str(criteria).lower()))
        criteria = AdvectionTimestepCriteria(lcfl=lcfl, Finf=Finf, gradFinf=gradFinf,
            parameter=parameter, criteria=criteria, 
            name=name, pretty_name=pretty_name, **kwds)
        self._push_criteria(parameter.name, criteria)
        return criteria.dt

    def push_stretching_criteria(self, criteria, gradFinf, cst=1.0,
            name=None, pretty_name=None, parameter=None, quiet=False, **kwds):
        """
        See hysop.operator.adapt_timpestep.StretchingTimestepCriteria.
        """
        parameter = self._build_parameter(parameter=parameter, quiet=quiet,
                name=name, pretty_name=pretty_name, basename='stretch')
        criteria = StretchingTimestepCriteria(cst=cst, parameter=parameter,
                gradFinf=gradFinf, criteria=criteria, **kwds)
        self._push_criteria(parameter.name, criteria)
        return criteria.dt

    def _build_parameter(self, name=None, pretty_name=None, quiet=None, 
                            basename=None, parameter=None):
        if (parameter is None):
            name         = first_not_None(name,'dt_{}'.format(basename))
            pretty_name  = first_not_None(pretty_name, name)
            parameter = ScalarParameter(name=name, pretty_name=pretty_name, 
                            quiet=quiet, dtype=HYSOP_REAL)
        return parameter

    def _push_criteria(self, name, criteria):
        assert not self.generated, 'Node was already generated.'
        if (name in self.criterias):
            msg='Criteria {} was already registered.'.format(name)
            raise RuntimeError(msg)
        self.criterias[name] = criteria

    def _generate(self):
        assert not self.generated, 'Node was already generated.'

        if not self.criterias:
            msg='No timestep criterias were ever setup.'
            raise RuntimeError(msg)

        # we need to compute the minimum timestep of all criterias
        # through a new operator.
        self.merge_kwds.setdefault('name', 'DT')
        merge = MergeTimeStepCriterias(parameter=self.parameter, 
                                       equivalent_CFL=self.equivalent_CFL,
                                       cfl_criteria=self.cfl_criteria,
                                       criterias=self.criterias, **self.merge_kwds)

        operators = self.criterias.values()
        operators.append(merge)
        return operators
