
from hysop.constants import Implementation
from hysop.fields.continuous_field import Field
from hysop.tools.types import check_instance, first_not_None, to_tuple
from hysop.tools.decorators import debug
from hysop.core.graph.computational_node_frontend import ComputationalGraphNodeFrontend
from hysop.topology.cartesian_descriptor import CartesianTopologyDescriptors

class Curl(ComputationalGraphNodeFrontend):
    """Generate an operator to compute the curl of a Field."""

    @classmethod
    def implementations(cls):
        from hysop.backend.host.python.operator.curl import PythonSpectralCurl
        from hysop.backend.device.opencl.operator.curl import OpenClSpectralCurl
        __implementations = {
                Implementation.PYTHON: PythonSpectralCurl,
                Implementation.OPENCL: OpenClSpectralCurl,
        }
        return __implementations
    
    @classmethod
    def default_implementation(cls):
        return Implementation.PYTHON

    @debug
    def __init__(self, Fin, Fout, variables, 
                    implementation=None, base_kwds=None, **kwds):
        """
        Create an operator that computes the curl of an input field Fin.

        Given Fin, a 2D ScalarField or VectorField or a 3D VectorField, compute Fout = curl(Fin).
        
        Only the following configurations are supported:
                 dim   nb_components  |   dim   nb_components
        Input:    2        (1,2)      |    3          3
        Output:   2        (2,1)      |    3          3

        Parameters
        ----------
        Fin: hysop.field.continuous_field.Field
            Continuous field as input ScalarField or VectorField.
            All contained field have to live on the same domain.
        Fout: hysop.field.continuous_field.Field
            Continuous field as output VectorField.
            All contained field have to live on the same domain.
        variables: dict
            dictionary of fields as keys and topologies as values.
        implementation: Implementation, optional, defaults to None
            target implementation, should be contained in available_implementations().
            If None, implementation will be set to default_implementation().
        kwds: dict, optional
            Extra parameters passed towards base class (MultiSpaceDerivatives).
        """
        base_kwds = first_not_None(base_kwds, {})
        check_instance(Fin, Field)
        check_instance(Fout, Field)
        check_instance(variables, dict, keys=Field, values=CartesianTopologyDescriptors)

        super(Curl, self).__init__(Fin=Fin, Fout=Fout, variables=variables,
                candidate_input_tensors=(Fin,),
                candidate_output_tensors=(Fout,),
                implementation=implementation, 
                base_kwds=base_kwds, **kwds)

