"""
@file min_max.py
MinMaxFieldStatistics: compute min(f), max(f) and/or max(|f|) for a given field f.
MinMaxDerivativeStatistics: compute min(d^k(Fi)/dXj^k), max(d^kFi/dXj^k) and/or max(|dFi/dXj|)
                            for a given field, component, direction and order.
MinMaxGradientStatistics: compute min(dFi/dXj), max(dFi/dXj) and/or max(|dFi/dXj|) 
                          for a given field, up to all components in all directions.
"""
from hysop import vprint
from hysop.constants import Backend, Implementation
from hysop.tools.types       import check_instance, first_not_None, to_tuple
from hysop.tools.enum        import EnumFactory
from hysop.tools.decorators  import debug
from hysop.tools.numpywrappers import npw
from hysop.tools.sympy_utils import nabla, partial, subscript
from hysop.fields.continuous_field import Field
from hysop.topology.cartesian_descriptor import CartesianTopologyDescriptors
from hysop.parameters.scalar_parameter import ScalarParameter
from hysop.parameters.tensor_parameter import TensorParameter
from hysop.core.graph.computational_operator import ComputationalGraphOperator
from hysop.core.graph.computational_node_frontend import ComputationalGraphNodeFrontend
from hysop.core.graph.graph import op_apply
from hysop.operator.derivative import Gradient
from hysop.operator.base.min_max import MinMaxFieldStatisticsBase, MinMaxDerivativeStatisticsBase


class MinMaxFieldStatistics(ComputationalGraphNodeFrontend):
    """
    Operator frontend to compute min and max statistics on the specific field.
    """
    
    @classmethod
    def implementations(cls):
        from hysop.backend.host.python.operator.min_max import PythonMinMaxFieldStatistics
        from hysop.backend.device.opencl.operator.min_max import OpenClMinMaxFieldStatistics
        implementations = {
                Implementation.PYTHON: PythonMinMaxFieldStatistics,
                Implementation.OPENCL: OpenClMinMaxFieldStatistics
        }
        return implementations
    
    @classmethod
    def default_implementation(cls):
        return Implementation.PYTHON
    
    @debug
    def __init__(self, field, components=None, coeffs=None,
            Fmin=None, Fmax=None, Finf=None, all_quiet=False,
            name=None, pbasename=None, ppbasename=None,
            variables=None, implementation=None, base_kwds=None, **kwds):
        """
        Initialize a MinMaxFieldStatistics operator frontend.
        Available operator backends are PYTHON and OPENCL.
        
        MinMaxFieldStatistics can compute some commonly required Field statistics:
            Fmin:  component-wise min values of the field.
            Fmax:  component-wise max values of the field.
            Finf:  component-wise max values of the absolute value of the field (computed using
                                                                                 Fmin and Fmax).

        All statistics are only computed if explicitely requested by user,
          unless required to compute another user-required statistic, see Notes.
        All statistics may also be additionaly scaled by a coefficient. 
        
        Compute vectorized statistics:
            Fmin = Smin * min(F[components])
            Fmax = Smax * max(F[components])
            Finf = Sinf * max(|Fmin|, |Fmax|)
        
        where F is an input field
              Fmin = created or supplied TensorParameter.
              Fmax = created or supplied TensorParameter.
              Finf = created or supplied TensorParameter.
              Smin = coeffs['Fmin']
              Smax = coeffs['Fmax']
              Sinf = coeffs['Finf']
        
        Parameters
        ----------
        field: Field
            The continuous field on which the statistics will be computed.
        components: array like of ints, optional
            The components on which the statistics are computed,
            defaults to all components.
        coeffs: dict of array like of coefficients, optional
            Optional scaling of the statistics.
            Scaling factor should be a scalar or an array-like of scalars for
            each components. If not given, defaults to 1 for all statistics.
        F...: TensorParameter or boolean, optional
            At least one statistic should be specified (either by boolean or TensorParameter).
            TensorParameters should be of shape (nb_components,), see Notes.
            If set to True, the TensorParameter will be generated automatically.
            Autogenerated TensorParameters that are not required by the user
            (ie. left to None or False during __init__) are, if required,
            generated but set to be quiet.
        all_quiet: bool
            Set all autogenerated TensorParameter to be quiet.
        name: str, optional
            Name of this operator.
        pbasename: str, optional
            Parameters basename for created parameters.
            Defaults to field.name.
        ppbasename: str, optional
            Parameters pretty basename for created parameters.
            Defaults to pbasename.
        variables: dict, optional
            Dictionary of fields as keys and topologies as values.
        implementation: hysop.constants.Implementation, optional
            Specify underlying backend implementation.
            Target implementation, should be contained in available_implementations().
            If None, implementation will be set to default_implementation().
        base_kwds: dict, optional
            Base class keyword arguments as a dictionnary.
        kwds: 
            Extra keyword arguments passed towards operator backend implementation.

        Attributes:
        -----------
        F...: TensorParameter
            All generated tensor parameters.
            Unused statistics are set to None.

        Notes
        -----
        nb_components = min(field.nb_components, len(components)).
        
        About statistics:
            Finf requires to compute Fmin and Fmax and will have value:
                Finf = Sinf * max( abs(Smin*Fmin), abs(Smax*Fmax))
            where Sinf, Smin and Smax are the scaling coefficients defined in coeffs.
        """
        check_instance(field, Field)
        check_instance(components, (tuple,list,npw.ndarray), values=int, 
                allow_none=True, min_value=0, max_value=field.nb_components-1)
        check_instance(coeffs, dict, keys=str, values=(tuple,list,npw.ndarray), allow_none=True)
        check_instance(variables, dict, keys=Field, 
                values=CartesianTopologyDescriptors, allow_none=True)
        check_instance(name, str, allow_none=True)
        check_instance(pbasename, str, allow_none=True)
        check_instance(ppbasename, (str, unicode), allow_none=True)
        check_instance(implementation, Implementation, allow_none=True)
        check_instance(base_kwds, dict, keys=str, allow_none=True)
        
        if (    ((Fmin is None) or (Fmin is False)) 
            and ((Fmax is None) or (Fmax is False))
            and ((Finf is None) or (Finf is False))):
            msg='No statistics were requested.'
            msg+='\nPlease specify Fmin, Fmax and/or Finf by either setting '
            msg+=' their value to True, or by by passing an already existing '
            msg+=' tensor parameter.'
            raise ValueError(msg)
            
        # Pregenerate parameters so that we can directly store them in self.
        parameters = MinMaxFieldStatisticsBase.build_parameters(field=field, 
                components=components, all_quiet=all_quiet,
                Fmin=Fmin, Fmax=Fmax, Finf=Finf, 
                pbasename=pbasename, ppbasename=ppbasename)

        (Fmin, Fmax, Finf) = tuple(parameters[k] for k in ('Fmin', 'Fmax', 'Finf'))

        super(MinMaxFieldStatistics, self).__init__(field=field, components=components,
                coeffs=coeffs, Fmin=Fmin, Fmax=Fmax, Finf=Finf, all_quiet=all_quiet,
                name=name, pbasename=pbasename, ppbasename=ppbasename,
                variables=variables, implementation=implementation, base_kwds=base_kwds,
                **kwds)

        self.Fmin, self.Fmax, self.Finf = (Fmin, Fmax, Finf)



class MinMaxDerivativeStatistics(ComputationalGraphNodeFrontend):
    """
    Operator frontend to compute min and max statistics on a specific 
    derivative of a field component, without keeping its output.
    """
    
    @classmethod
    def implementations(cls):
        raise NotImplementedError

    @classmethod
    def default_implementation(cls):
        raise NotImplementedError
    
    @debug
    def __init__(self, F, dF=None, A=None, 
            derivative=None, component=None, direction=None,
            out_component=None, scaling_view=None, 
            Fmin=None, Fmax=None, Finf=None, coeffs=None, all_quiet=False,
            name=None, pbasename=None, ppbasename=None,
            variables=None, implementation=None, base_kwds=None, **kwds):
        """
        Initialize a MinMaxDerivativeStatistics operator frontend.
        Available operator backends are PYTHON and OPENCL.
        
        MinMaxDerivativeStatistics can compute some commonly required Field derivative statistics:
            Fmin: min value of a derivative of the field.
            Fmax: max value of a derivative of the field.
            Finf: max value of the absolute value of a
                    derivative of the field (computed using Fmin and Fmax).
        
        First compute the derivative of a component of a field F in a given direction
        at a given order and on a given backend out of place in a specific output component of 
        dF. The derivative is then possibly scaled by another field/parameter/value A.

        After the scaled derivative has been computed, compute user requested statistics
        (min and max values) on this new field and scale those statistics by other scaling 
        parameters stored in coeffs.

        1) Compute derivative
            dF[k] = alpha * d^n(F[i])/dXj**n

        2) Compute statistics
            Fmin = Smin * min(dF[k])
            Fmax = Smax * max(dF[k])
            Finf = Sinf * max(|Fmin|, |Fmax|)

        where F  is an input field
              dF is an output field (by default a temporary field).
              k = out_component < dF.nb_components
              i =     component <  F.nb_components
              n = derivative order > 0
              alpha = A[scaling_view]
                where A is a Field, a Parameter or a scalar.
                      scaling_view is a component, a slice or None.
              Fmin = created or supplied TensorParameter.
              Fmax = created or supplied TensorParameter.
              Finf = created or supplied TensorParameter.
              Smin = coeffs['Fmin']
              Smax = coeffs['Fmax']
              Sinf = coeffs['Finf']

        Statistics are only computed if explicitely requested by user,
          unless required to compute another user-required statistic, see Notes.
        
        Parameters
        ----------
        F: hysop.field.continuous_field.Field
            Continuous field as input.
        dF: hysop.field.continuous_field.Field, optional
            Continuous field to be written.
            Some backend may allow inplace differentiation.
            By default a temporary scalar field is created,
            which means that the computation of the derivative
            may be discarded after this operator has been applied
            because of temporary buffer sharing between operators.
        A: numerical value, ScalarParameter or Field, optional
            Scaling field/parameter/value for convenience.
            Defaults to no scaling.
        derivative: int, optional
            Which derivative to generate.
            Defaults to 1.
        component: int, optional
            The component on which to take the derivative.
            Defaults to 0.
        direction: int, optional
            Directions in which to take the derivative.
            Defaults to 0.
        out_component: int, optional
            The component were the result will be stored.
            Defaults to component.
        scaling_view: int or slice, optional
            View on the scaling field/parameter/value A.
            Should be a component if this is a field.
            Should be a slice if this is a TensorParameter.
            Should be None it this is a ScalarParameter or
            a numerical value.
            Should only be given if A is given.
        F...: TensorParameter or boolean, optional
            The output parameters that will contain the statistics.
            At least one statistic should be specified (either by boolean or TensorParameter).
            TensorParameters should be of shape (1,).
            If set to True, the TensorParameter will be generated automatically.
            Autogenerated TensorParameters that are not required by the user
            (ie. left to None or False during __init__) are, if required,
            generated but set to be quiet.
            Autogenerated parameters can be retrieved using the 'Fmin', 'Fmax'
            and 'Finf' attributes.
        all_quiet: bool
            Set all autogenerated TensorParameter to be quiet.
        coeffs: dict of array like of coefficients, optional
            Optional scaling of the statistics.
            Scaling factor should be a scalar or an array-like of scalars for
            each components. If not given, defaults to 1 for all statistics.
        name: str, optional
            Name of this operator.
        pbasename: str, optional
            Parameters basename for created parameters.
            Defaults to field.name.
        ppbasename: str, optional
            Parameters pretty basename for created parameters.
            Defaults to pbasename.
        variables: dict, optional
            Dictionary of fields as keys and topologies as values.
        implementation: hysop.constants.Implementation, optional
            Specify underlying backend implementation.
            Target implementation, should be contained in available_implementations().
            If None, implementation will be set to default_implementation().
        base_kwds: dict, optional
            Base class keyword arguments as a dictionnary.
        kwds: 
            Extra keyword arguments passed towards operator backend implementation.

        Attributes:
        -----------
        F...: TensorParameter
            All generated tensor parameters.
            Unused statistics are set to None.

        Notes
        -----
        About statistics:
            Finf requires to compute Fmin and Fmax and will have value:
                Finf = Sinf * max( abs(Smin*Fmin), abs(Smax*Fmax))
            where Sinf, Smin and Smax are the scaling coefficients defined in coeffs.
        """
        check_instance(F, Field)
        check_instance(dF, Field, allow_none=True)
        check_instance(derivative, int, allow_none=True)
        check_instance(component, int, allow_none=True)
        check_instance(direction, int, allow_none=True)
        check_instance(out_component, int, allow_none=True)
        check_instance(coeffs, dict, keys=str, values=(int, float, npw.number), allow_none=True)
        check_instance(variables, dict, keys=Field, values=CartesianTopologyDescriptors,
                        allow_none=True)
        check_instance(name, str, allow_none=True)
        check_instance(pbasename, str, allow_none=True)
        check_instance(ppbasename, (str, unicode), allow_none=True)
        check_instance(implementation, Implementation, allow_none=True)
        check_instance(base_kwds, dict, keys=str, allow_none=True)
        
        if (    ((Fmin is None) or (Fmin is False)) 
            and ((Fmax is None) or (Fmax is False))
            and ((Finf is None) or (Finf is False))):
            msg='No statistics were requested.'
            msg+='\nPlease specify Fmin, Fmax and/or Finf by either setting '
            msg+=' their value to True, or by by passing an already existing '
            msg+=' tensor parameter.'
            raise ValueError(msg)

        # Pregenerate parameters so that we can directly store them in self.
        parameters = MinMaxDerivativeStatisticsBase.build_parameters(field=F, 
                components=(component,), all_quiet=all_quiet,
                Fmin=Fmin, Fmax=Fmax, Finf=Finf, 
                pbasename=pbasename, ppbasename=ppbasename)
        (Fmin, Fmax, Finf) = tuple(parameters[k] for k in ('Fmin', 'Fmax', 'Finf'))

        super(MinMaxDerivativeStatistics, self).__init__(F=F, dF=dF, A=A, 
            derivative=derivative, component=component, direction=direction,
            out_component=out_component, scaling_view=scaling_view, 
            Fmin=Fmin, Fmax=Fmax, Finf=Finf, coeffs=coeffs, all_quiet=all_quiet,
            name=name, pbasename=pbasename, ppbasename=ppbasename,
            variables=variables, implementation=implementation, 
            base_kwds=base_kwds, **kwds)
        
        self.Fmin, self.Fmax, self.Finf = (Fmin, Fmax, Finf)


class MinMaxSpectralDerivativeStatistics(MinMaxDerivativeStatistics):
    """
    Operator frontend to compute min and max statistics on a specific 
    derivative of a field component using the spectral method.
    """
    @classmethod
    def implementations(cls):
        from hysop.backend.host.python.operator.min_max import \
                PythonMinMaxSpectralDerivativeStatistics
        from hysop.backend.device.opencl.operator.min_max import \
                OpenClMinMaxSpectralDerivativeStatistics
        implementations = {
                Implementation.PYTHON: PythonMinMaxSpectralDerivativeStatistics,
                Implementation.OPENCL: OpenClMinMaxSpectralDerivativeStatistics
        }
        return implementations

    @classmethod
    def default_implementation(cls):
        return Implementation.PYTHON


class MinMaxFiniteDifferencesDerivativeStatistics(MinMaxDerivativeStatistics):
    """
    Operator frontend to compute min and max statistics on a specific 
    derivative of a field component using finite differences.
    """
    @classmethod
    def implementations(cls):
        from hysop.backend.host.python.operator.min_max import \
                PythonMinMaxFiniteDifferencesDerivativeStatistics
        from hysop.backend.device.opencl.operator.min_max import \
                OpenClMinMaxFiniteDifferencesDerivativeStatistics
        implementations = {
                Implementation.PYTHON: PythonMinMaxFiniteDifferencesDerivativeStatistics,
                Implementation.OPENCL: OpenClMinMaxFiniteDifferencesDerivativeStatistics
        }
        return implementations

    @classmethod
    def default_implementation(cls):
        return Implementation.PYTHON


class MinMaxGradientStatistics(Gradient):
    """
    Interface for computing some statistics on the gradient of a field (minimum and maximum)
    one component at a time to limit memory usage.
    Generate multiple MinMaxDerivativeStatistics operators.
    """
    
    @debug
    def __init__(self, F, gradF=None, directions=None, coeffs=None,
            Fmin=None, Fmax=None, Finf=None, 
            all_quiet=True, print_tensors=True,
            name=None, pretty_name=None, pbasename=None, ppbasename=None,
            variables=None, implementation=None, base_kwds=None, 
            cls=MinMaxFiniteDifferencesDerivativeStatistics,
            **kwds):
        """
        Create an operator generator that yields a sequence of operators
        that compute statistics on the gradient of an input field F.
        
        MinMaxGradientStatistics can compute some commonly used Field statistics:
            Fmin:  component-wise and direction-wise min values of the gradient of the field.
            Fmax:  component-wise and direction-wise max values of the gradient of the field.
            Finf:  component-wise and direction-wise max values of the absolute value of the 
                        gradient of the field (computed using Fmin and Fmax).

        Derivatives can be computed with respect to specific directions and not necessarily in 
        all directions. To restrict the number of components, take a tensor view on F (and gradF).
        
        ----------------------------------------------
        Let k = idx + (j,)
        gradF[k] = dF[idx]/dXd
        ----------------------------------------------
        Fmax[k]  = Smin * min( dF[idx]/dXd )
        Fmin[k]  = Smax * max( dF[idx]/dXd )
        Finf[k]  = Sinf * max( |Fmin[k]|, |Fmax[k]| )
        ----------------------------------------------
        where F is an input field,
              nb_components = F.nb_components = np.prod(F.shape)
              nb_directions = min( F.dim, len(directions))
              gradF is an optional output tensor field,
              idx is contained in numpy.ndindex(*F.shape)
              0 <= j < nb_directions
              d = directions[j]
              Fmin = created or supplied TensorParameter of shape F.shape + (nb_directions,).
              Fmax = created or supplied TensorParameter of shape F.shape + (nb_directions,).
              Finf = created or supplied TensorParameter of shape F.shape + (nb_directions,).
              Smin = coeffs['Fmin'] or coeffs['Fmin'][k]
              Smax = coeffs['Fmax'] or coeffs['Fmax'][k]
              Sinf = coeffs['Finf'] or coeffs['Fmax'][k]

        All statistics are only computed if explicitely required by user,
          unless required to compute another required statistic, see Notes.
        
        Parameters
        ----------
        F: Field
            The continuous input field on which the gradient
            will be taken and statistics will be computed.
        gradF: Field, optional
            Optional output field for the gradient.
            If the gradient is required as an output, one can also use MinMaxStatistics
            on a precomputed gradient (using the Gradient operator) instead of 
            MinMaxGradientStatistics.
        directions: array like of ints, optional
            The directions in which the statistics are computed,
            defaults to all directions (ie. range(F.dim)).
        coeffs: dict of scalar or array like of coefficients, optional
            Optional scaling of the statistics.
            Scaling factor should be a scalar or an array-like of scalars for
            each components of gradF. If not given default to 1 for all statistics.
        F...: TensorParameter or boolean, optional
            At least one statistic should be specified (either by boolean or TensorParameter).
            TensorParameters should be of shape F.shape + (nb_directions,), see Notes.
            If set to True, the TensorParameter will be generated automatically.
            Autogenerated TensorParameters that are not required by the user
            but are generated anyway are set to be quiet.
            All autogenerated TensorParameters can be retrieved as attributes
            of this object.
        all_quiet: bool, optional, defaults to True
            Set all generated params to be quiet, even the ones that are requested
            explicitely. 
        print_tensors: bool, optional, defaults to True
            Should the phony operator print the tensor parameters during apply ?
        name: str, optional
            Name template for generated operator names.
            Defaults to MinMax({}) where {} will be replaced by gradF[k].name.
        pretty_name: str, optional
            Name template for generaed operatr pretty names.
            Defaults to |+/-{}| where {} will be replaced by gradF[k].pretty_name.
        pbasename: str, optional
            Basename for created tensor parameters.
            Defaults to gradF.name.
        ppbasename: str, optional
            Pretty basename for created tensor parameters.
            Defaults to gradF.pretty_name.
        variables: dict
            Dictionary of fields as keys and topologies as values.
        implementation: hysop.constants.Implementation, optional
            Specify generated operator underlying backend implementation.
            Target implementation, should be contained in 
              MinMaxDerivativeStatistics.available_implementations().
            If None, implementation will be set to 
              MinMaxDerivativeStatistics.default_implementation().
        base_kwds: dict
            Base class keyword arguments.
        kwds: dict
            Additional kwds passed to chosen implementation.

        Attributes:
        -----------
        Fmin, Fmax, Finf: TensorParameter
            All generated tensor parameters.
            Unused statistics are set to None.

        Notes
        -----
        nb_components = F.nb_components
        nb_directions = min(F.dim, len(directions)).
        
        About statistics:
            Finf requires to compute Fmin and Fmax.
                Finf = Sinf * max( abs(Smin*Fmin), abs(Smax*Fmax))
            where Sinf, Smin and Smax are the scaling coefficients defined in coeffs.
        """

        check_instance(F, Field)
        check_instance(gradF, Field, allow_none=True)
        check_instance(directions, tuple, values=int, allow_none=True, 
                minval=0, maxval=F.dim-1, minsize=1, unique=True)
        check_instance(coeffs, dict, keys=str, values=(int, float, npw.number), allow_none=True)
        check_instance(variables, dict, keys=Field, values=CartesianTopologyDescriptors, 
                        allow_none=True)
        check_instance(name, str, allow_none=True)
        check_instance(pbasename, str, allow_none=True)
        check_instance(ppbasename, (str,unicode), allow_none=True)
        check_instance(implementation, Implementation, allow_none=True)
        check_instance(base_kwds, dict, allow_none=True)
        check_instance(all_quiet, bool, allow_none=True)

        if (    ((Fmin is None) or (Fmin is False)) 
            and ((Fmax is None) or (Fmax is False))
            and ((Finf is None) or (Finf is False))):
            msg='No statistics were requested.'
            msg+='\nPlease specify Fmin, Fmax and/or Finf by either setting '
            msg+=' their value to True, or by by passing an already existing '
            msg+=' tensor parameter.'
            raise ValueError(msg)
        
        coeffs     = first_not_None(coeffs, {})
        variables  = first_not_None(variables, {F: None})
        all_quiet  = first_not_None(all_quiet, False)
        
        directions = to_tuple(first_not_None(directions, range(F.dim)))
        nb_directions = len(directions)
        
        if F.is_tensor:
            oshape = F.shape + (nb_directions,)
        else:
            oshape = (nb_directions,)
        
        if (gradF is None):
            gradF = F.gradient(directions=directions, is_tmp=True)
        assert (gradF.shape == oshape), gradF.shape
        
        variables.setdefault(gradF, variables[F])
        
        _names = {
            'Fmin': '{}_min',
            'Fmax': '{}_max',
            'Finf': '{}_inf'
        }

        _pretty_names = {
            'Fmin': u'{}\u208b',
            'Fmax': u'{}\u208a',
            'Finf': u'|{}|\u208a'
        }

        pbasename  = first_not_None(pbasename, gradF.name)
        ppbasename = first_not_None(ppbasename, gradF.pretty_name)
        
        names = { k: v.format(pbasename) for (k,v) in _names.iteritems() }
        pretty_names = { k: v.format(ppbasename.decode('utf-8')) 
                            for (k,v) in _pretty_names.iteritems() }

        def make_param(k, quiet):
            return TensorParameter(name=names[k], pretty_name=pretty_names[k],
                    dtype=F.dtype, shape=oshape, quiet=quiet)

        parameters = {}
        _parameters = dict(Fmin=Fmin, Fmax=Fmax, Finf=Finf)
        for (k,v) in _parameters.iteritems():
            param = _parameters[k]
            if isinstance(param, TensorParameter):
                pass
            elif (param is True):
                param = make_param(k, quiet=all_quiet)
            elif (_parameters['Finf'] is not None) and ((k == 'Fmin') or (k == 'Fmax')):
                param = make_param(k, quiet=True)
            else:
                param = None
                setattr(self, k, param)
                continue
            assert param.shape == oshape
            coeffs.setdefault(k, 1)
            parameters[k] = param
            setattr(self, k, param)

        unused_coeffs = set(coeffs.keys()) - set(parameters.keys())
        if unused_coeffs:
            msg='The following coefficients are not needed: {}'
            msg=msg.format(unused_coeffs)
            raise ValueError(unused_coeffs)
        
        name = first_not_None(name, 'MinMax({})')
        pretty_name = first_not_None(pretty_name, u'|\u00b1{}|')

        extra_params = { 'name': gradF.new_empty_array(), 
                         'pretty_name': gradF.new_empty_array(), 
                         'coeffs': coeffs, 
                         'implementation': implementation }

        for (idx, Fi) in gradF.nd_iter():
            for (statname, stat) in parameters.iteritems():
                if (stat is None):
                    continue
                pname  = _names[statname].format(Fi.name)
                ppname = _pretty_names[statname].format(Fi.pretty_name.decode('utf-8'))
                S = stat.view(idx=idx, name=pname, pretty_name=ppname)
                stats = extra_params.setdefault(statname, gradF.new_empty_array())
                stats[idx] = S
            extra_params['name'][idx] = name.format(Fi.name)
            extra_params['pretty_name'][idx] = pretty_name.format(
                                                Fi.pretty_name.decode('utf-8')).encode('utf-8')

        super(MinMaxGradientStatistics, self).__init__(F=F, gradF=gradF,
                directions=directions, extra_params=extra_params,
                cls=cls, variables=variables, **kwds)

        # add a phony operator to gather parameter views
        class MergeTensorViewsOperator(ComputationalGraphOperator):
            @op_apply
            def apply(self, **kwds):
                super(MergeTensorViewsOperator, self).apply(**kwds)
                if not print_tensors:
                    return
                for (k,v) in _parameters.iteritems():
                    if (v is not None) and (v is not False):
                        param = parameters[k]
                        msg='>Parameter {} set to:\n{}'.format(param.pretty_name, param.value)
                        vprint(msg)


        _phony_input_params = {}
        _phony_output_params = { }
        for pname in _names.keys():
            if (pname in extra_params):
                param = parameters[pname]
                _phony_input_params.update({p.name:p for p in extra_params[pname].ravel()})
                _phony_output_params[param.name] = param
        op = MergeTensorViewsOperator(name=name.format(gradF.name), 
                pretty_name=pretty_name.format(gradF.pretty_name.decode('utf-8')),
                input_params=_phony_input_params, 
                output_params=_phony_output_params)
        self._phony_op = op

    @debug
    def _generate(self):
        """Generate all operators."""
        operators = super(MinMaxGradientStatistics, self)._generate()
        operators += (self._phony_op,)
        return operators
    
    @debug
    def generate_direction(self, i, dt_coeff):
        # See MultiSpaceDerivatives for the directional interface
        ops = super(MinMaxGradientStatistics, self).generate_direction(i=i, dt_coeff=dt_coeff)
        if ops and (i==max(self.directions)):
            ops += (self._phony_op,)
        return ops


